import json
from enum import Enum


class MessageType(Enum):
    SYNC_MESSAGE = 1
    DATA_MESSAGE = 2


class Message:
    def __init__(
        self,
        source: str,
        timestamp: int,
        type: MessageType,
        text: str,
        group: str = None,
        reaction: str = None,
        raw_message: str = None,
    ):
        # required
        self.source = source
        self.timestamp = timestamp
        self.type = type
        self.text = text
        # optional
        self.group = group
        self.reaction = reaction

        self.raw_message = raw_message

    @classmethod
    def parse(cls, raw_message: str):
        try:
            raw_message = json.loads(raw_message)
        except Exception:
            raise UnknownMessageFormatError

        # General attributes
        try:
            source = raw_message["envelope"]["source"]
            timestamp = raw_message["envelope"]["timestamp"]
        except Exception:
            raise UnknownMessageFormatError

        # Option 1: syncMessage
        if "syncMessage" in raw_message["envelope"]:
            type = MessageType.SYNC_MESSAGE
            text = cls._parse_sync_message(raw_message["envelope"]["syncMessage"])
            group = cls._parse_group_information(
                raw_message["envelope"]["syncMessage"]["sentMessage"]
            )
            reaction = cls._parse_reaction(
                raw_message["envelope"]["syncMessage"]["sentMessage"]
            )

        # Option 2: dataMessage
        elif "dataMessage" in raw_message["envelope"]:
            type = MessageType.DATA_MESSAGE
            text = cls._parse_data_message(raw_message["envelope"]["dataMessage"])
            group = cls._parse_group_information(raw_message["envelope"]["dataMessage"])
            reaction = cls._parse_reaction(raw_message["envelope"]["dataMessage"])

        else:
            raise UnknownMessageFormatError

        return cls(source, timestamp, type, text, group, reaction)

    @classmethod
    def _parse_sync_message(cls, sync_message: dict) -> str:
        try:
            text = sync_message["sentMessage"]["message"]
            return text
        except Exception:
            raise UnknownMessageFormatError

    @classmethod
    def _parse_data_message(cls, data_message: dict) -> str:
        try:
            text = data_message["message"]
            return text
        except Exception:
            raise UnknownMessageFormatError

    @classmethod
    def _parse_group_information(self, message: dict) -> str:
        try:
            group = message["groupInfo"]["groupId"]
            return group
        except Exception:
            return None

    @classmethod
    def _parse_reaction(self, message: dict) -> str:
        try:
            reaction = message["reaction"]["emoji"]
            return reaction
        except Exception:
            return None

    def __str__(self):
        if self.text is None:
            return ""
        return self.text


class UnknownMessageFormatError(Exception):
    pass



import redis
import json
from typing import Any


class Storage:
    def exists(self, key: str) -> bool:
        raise NotImplementedError

    def read(self, key: str) -> Any:
        raise NotImplementedError

    def save(self, key: str, object: Any):
        raise NotImplementedError


class StorageError(Exception):
    pass


class InMemoryStorage(Storage):
    def __init__(self):
        self._storage = {}

    def exists(self, key: str) -> bool:
        return key in self._storage

    def read(self, key: str) -> Any:
        try:
            result_str = self._storage[key]
            result_dict = json.loads(result_str)
            return result_dict
        except Exception as e:
            raise StorageError(f"InMemory load failed: {e}")

    def save(self, key: str, object: Any):
        try:
            object_str = json.dumps(object)
            self._storage[key] = object_str
        except Exception as e:
            raise StorageError(f"InMemory save failed: {e}")


class RedisStorage(Storage):
    def __init__(self, host, port):
        self._redis = redis.Redis(host=host, port=port, db=0)

    def exists(self, key: str) -> bool:
        return self._redis.exists(key)

    def read(self, key: str) -> Any:
        try:
            result_bytes = self._redis.get(key)
            result_str = result_bytes.decode("utf-8")
            result_dict = json.loads(result_str)
            return result_dict
        except Exception as e:
            raise StorageError(f"Redis load failed: {e}")

    def save(self, key: str, object: Any):
        try:
            object_str = json.dumps(object)
            self._redis.set(key, object_str)
        except Exception as e:
            raise StorageError(f"Redis save failed: {e}")



import unittest
import uuid
import time
import json
from ..bot import SignalBot


class ChatTestCase(unittest.IsolatedAsyncioTestCase):
    signal_service = "127.0.0.1:8080"
    phone_number = "+49123456789"

    group_id = "group_id1"
    group_secret = "group.group_secret1"
    config = {
        "signal_service": signal_service,
        "phone_number": phone_number,
    }

    def setUp(self):
        self.signal_bot = SignalBot(ChatTestCase.config)
        self.signal_bot.listen(ChatTestCase.group_id, ChatTestCase.group_secret)

    async def run_bot(self):
        PRODUCER_ID = 1337
        HANDLER_ID = 4444
        await self.signal_bot._produce(PRODUCER_ID)
        while self.signal_bot._q.qsize() > 0:
            await self.signal_bot._consume_new_item(HANDLER_ID)

    @classmethod
    def new_message(cls, text) -> str:
        timestamp = time.time()
        new_uuid = str(uuid.uuid4())
        message = {
            "envelope": {
                "source": ChatTestCase.phone_number,
                "sourceNumber": ChatTestCase.phone_number,
                "sourceUuid": new_uuid,
                "sourceName": "some_source_name",
                "sourceDevice": 1,
                "timestamp": timestamp,
                "syncMessage": {
                    "sentMessage": {
                        "timestamp": timestamp,
                        "message": text,
                        "expiresInSeconds": 0,
                        "viewOnce": False,
                        "mentions": [],
                        "attachments": [],
                        "contacts": [],
                        "groupInfo": {
                            "groupId": ChatTestCase.group_id,
                            "type": "DELIVER",
                        },
                        "destination": None,
                        "destinationNumber": None,
                        "destinationUuid": None,
                    }
                },
            }
        }
        return json.dumps(message)



import asyncio
import time
from apscheduler.schedulers.asyncio import AsyncIOScheduler
import logging
from typing import Union


from .api import SignalAPI, ReceiveMessagesError
from .command import Command
from .message import Message, UnknownMessageFormatError
from .storage import RedisStorage, InMemoryStorage
from .context import Context


class SignalBot:
    def __init__(self, config: dict):
        """SignalBot

        Example Config:
        ===============
        signal_service: "127.0.0.1:8080"
        phone_number: "+49123456789"
        storage:
            redis_host: "redis"
            redis_port: 6379
        scheduler:
            enable_jobstore: true
        """
        self.config = config
        self.commands = []
        self.groups = {}

        # Signal API (required)
        self._init_api()

        # Event loop (required)
        self._init_event_loop()

        # Storage (optional)
        self._init_storage()

        # Scheduler settings (jobstore)
        self._init_scheduler()

    def _init_api(self):
        try:
            self._phone_number = self.config["phone_number"]
            self._signal_service = self.config["signal_service"]
            self._signal = SignalAPI(self._signal_service, self._phone_number)
        except KeyError:
            raise SignalBotError("Could not initialize SignalAPI with given config")

    def _init_event_loop(self):
        self._event_loop = asyncio.get_event_loop()
        self._q = asyncio.Queue()

    def _init_storage(self):
        try:
            config_storage = self.config["storage"]
            self._redis_host = config_storage["redis_host"]
            self._redis_port = config_storage["redis_port"]
            self.storage = RedisStorage(self._redis_host, self._redis_port)
        except Exception:
            self.storage = InMemoryStorage()
            logging.warn(
                "[Bot] Could not initialize Redis. In-memory storage will be used. "
                "Restarting will delete the storage!"
            )

    def _init_scheduler(self):
        # required: scheduler
        try:
            self.scheduler = AsyncIOScheduler(event_loop=self._event_loop)
        except Exception as e:
            raise SignalBotError(f"Could not initialize scheduler: {e}")

        # optional: jobstore option
        # TODO: do not use jobstore because it does not work with coroutines
        # try:
        #     enable_jobstore = self.config["scheduler"]["enable_jobstore"]
        #     if enable_jobstore is False:
        #         return
        #     if not isinstance(self.storage, RedisStorage):
        #         raise Exception("Scheduler is not Redis Storage")
        #     host, port = self._redis_host, self._redis_port
        #     self.scheduler.add_jobstore(
        #         "redis",
        #         jobs_key="Signalbot:scheduler:jobs",
        #         run_times_key="Signalbot:scheduler:run_times",
        #         host=host,
        #         port=port,
        #     )
        # except Exception as e:
        #     logging.warn(
        #         "[Bot] Could not enable scheduler's job store, despite setting "
        #         f"`enable_jobstore` to True: {e}"
        #     )
        #     return

    def listen(self, group: str, secret: str):
        self.groups[group] = secret

    def register(self, command: Command):
        command.bot = self
        command.setup()
        self.commands.append(command)

    def start(self):
        self._event_loop.create_task(self._produce_consume_messages())

        # Add more scheduler tasks here
        # self.scheduler.add_job(...)
        self.scheduler.start()

        # Run event loop
        self._event_loop.run_forever()

    async def send(self, receiver: Union[Message, str], text: str):
        receiver_secret = self._resolve_receiver(receiver)
        await self._signal.send(receiver_secret, text)
        logging.info(f"[Bot] New message sent:\n{text}")

    async def react(self, message: Message, emoji: str):
        # TODO: check that emoji is really an emoji
        recipient = self._resolve_receiver(message)
        target_author = message.source
        timestamp = message.timestamp
        await self._signal.react(recipient, emoji, target_author, timestamp)
        logging.info(f"[Bot] New reaction: {emoji}")

    async def start_typing(self, receiver: Union[Message, str]):
        receiver_secret = self._resolve_receiver(receiver)
        await self._signal.start_typing(receiver_secret)

    async def stop_typing(self, receiver: Union[Message, str]):
        receiver_secret = self._resolve_receiver(receiver)
        await self._signal.stop_typing(receiver_secret)

    def _resolve_receiver(self, receiver: Union[Message, str]) -> str:
        # accept both messages and direct group ids
        if isinstance(receiver, Message):
            group_id = receiver.group
        else:
            group_id = receiver

        # resolve group id by using group secrets
        if group_id not in self.groups:
            raise SignalBotError("group_id is not in self.groups")
        receiver_secret = self.groups[group_id]

        return receiver_secret

    async def _produce_consume_messages(self, producers=1, consumers=3) -> None:
        producers = [
            asyncio.create_task(self._produce(n)) for n in range(1, producers + 1)
        ]
        consumers = [
            asyncio.create_task(self._consume(n)) for n in range(1, consumers + 1)
        ]
        await asyncio.gather(*producers)
        await self._q.join()
        for c in consumers:
            c.cancel()

    async def _produce(self, name: int) -> None:
        logging.info(f"[Bot] Producer #{name} started")
        try:
            async for raw_message in self._signal.receive():
                logging.info(f"[Raw Message] {raw_message}")

                try:
                    message = Message.parse(raw_message)
                except UnknownMessageFormatError:
                    continue

                group = message.group
                if group not in self.groups:
                    continue

                for command in self.commands:
                    await self._q.put((command, message, time.perf_counter()))

        except ReceiveMessagesError as e:
            # TODO: retry strategy
            raise SignalBotError(f"Cannot receive messages: {e}")

    async def _consume(self, name: int) -> None:
        logging.info(f"[Bot] Consumer #{name} started")
        while True:
            try:
                await self._consume_new_item(name)
            except Exception:
                continue

    async def _consume_new_item(self, name: int) -> None:
        command, message, t = await self._q.get()
        now = time.perf_counter()
        logging.info(f"[Bot] Consumer #{name} got new job in {now-t:0.5f} seconds")

        # handle Command
        try:
            context = Context(self, message)
            await command.handle(context)
        except Exception as e:
            logging.error(f"[{command.__class__.__name__}] Error: {e}")
            raise e

        # done
        self._q.task_done()


class SignalBotError(Exception):
    pass
